from asn1_prime import SAFEPRIME, GENERATOR
from asn1_key import PUBKEY, SECKEY

################################################
# El-gamal Signature implementation (DSA-like) #
################################################

import hashlib

def elgamal_verify(r, s, m):
    if r <= 0 or r >= SAFEPRIME:
        return False
    if s <= 0 or s >= SAFEPRIME-1:
        return False
    h = int(hashlib.sha384(m).hexdigest(), 16)
    left = pow(GENERATOR, h, SAFEPRIME)
    right = (pow(PUBKEY, r, SAFEPRIME) * pow(r, s, SAFEPRIME)) % SAFEPRIME
    return left == right


### Encoding

from pyasn1.type import univ, namedtype, tag

class ElgamalSig(univ.Sequence):
    componentType = namedtype.NamedTypes(
        namedtype.NamedType('r', univ.Integer()),
        namedtype.NamedType('s', univ.Integer()),
        namedtype.NamedType('m', univ.OctetString())
    )

###########
# Handler #
###########

DUPLICATES = []

def is_duplicate(s):
    return s in DUPLICATES

import base64, SocketServer, os, sys
from pyasn1.codec.ber import decoder

class ServerHandler(SocketServer.BaseRequestHandler):

    def fail(self, message):
        self.request.sendall(message + "\nGood-bye.\n")
        self.request.close()
        return False

    def captcha(self):
        proof = base64.b64encode(os.urandom(9))
        self.request.sendall(proof)
        test = self.request.recv(20)
        ha = hashlib.sha1()
        ha.update(test)
        if test[0:12]!=proof or not ha.digest().endswith('\xFF\xFF\xFF'):
            self.fail("You're a robot!")

    def handle(self):
        self.captcha()
        sig_raw = self.request.recv(5000)
        sig, _ = decoder.decode(sig_raw, asn1Spec=ElgamalSig())
        r = int(sig.getComponentByName('r'))
        s = int(sig.getComponentByName('s'))
        m = str(sig.getComponentByName('m'))
        if not elgamal_verify(r, s, m):
            self.request.close()
        elif is_duplicate(sig_raw):
            self.request.close()
        elif m != "There is no need to be upset":
            self.request.close()
        else:
            self.request.sendall(FLAG)
            self.request.close()

class ThreadedServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    pass


##########
# Server #
##########

import json

FLAG = ""

if __name__ == "__main__":
    HOST = sys.argv[1]
    PORT = int(sys.argv[2])

    FLAG = open('flag.txt', 'r').read()

    for fname in os.listdir('sigs'):
        with open('sigs/' + fname, 'r') as f:
            sig = f.read()
            DUPLICATES.append(sig)
    
    server = ThreadedServer((HOST, PORT), ServerHandler)
    server.allow_reuse_address = True
    server.serve_forever()
    
